/*
 Wireless Gameboy Controller Cart
 Version: 1.2a
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 23/09/2018
 Last Modified: 20/11/2019
 
 Use your Gameboy, GBC, GBP, GBA or GBA SP as a Wireless game controller for any device which supports USB HID Joysticks/Gamepads.
 This would allow you to play games on your PC with a Gameboy of your choosing.
 
 */

#include <stdio.h>
#include <gb/gb.h>
#include <gb/console.h>
#include <gb/drawing.h>

#include "bar_c.h"
#include "bar_c.c"
#include "bar_m.c"

void init();
void options();
void checkInput();
void timer_isr();
WORD senddata(WORD);
WORD updatenrf(WORD);
WORD updateppr(WORD);
WORD sleep();

UWORD kpRead = 0x00;
UWORD kpReadLast = 0x00;
UWORD kpInput = 0x00;
UBYTE refreshScreen = 1;
UBYTE transmitOptions = 0;
UBYTE inOptions = 0;
UBYTE arrowLocation = 1;
UBYTE screenOn = 1;
UBYTE showKeyPresses = 0;
UWORD rfChannel = 1;
UWORD rfAddress = 1;
UWORD rfPpr = 1;

const UWORD bar_p[] = {
  bar_cCGBPal0c0,bar_cCGBPal0c1,bar_cCGBPal0c2,bar_cCGBPal0c3,
  bar_cCGBPal1c0,bar_cCGBPal1c1,bar_cCGBPal1c2,bar_cCGBPal1c3,
  bar_cCGBPal2c0,bar_cCGBPal2c1,bar_cCGBPal2c2,bar_cCGBPal2c3,
  bar_cCGBPal3c0,bar_cCGBPal3c1,bar_cCGBPal3c2,bar_cCGBPal3c3,
  bar_cCGBPal4c0,bar_cCGBPal4c1,bar_cCGBPal4c2,bar_cCGBPal4c3,
  bar_cCGBPal5c0,bar_cCGBPal5c1,bar_cCGBPal5c2,bar_cCGBPal5c3,
  bar_cCGBPal6c0,bar_cCGBPal6c1,bar_cCGBPal6c2,bar_cCGBPal6c3,
  bar_cCGBPal7c0,bar_cCGBPal7c1,bar_cCGBPal7c2,bar_cCGBPal7c3
};


void main() {
	// Select button held down at boot - Screen off
	if (joypad() & J_SELECT) {
		screenOn = 0;
		DISPLAY_OFF;
		
		// If down key also held down, show black background instead for backlit screens so they don't show white
		if (joypad() & J_DOWN) {
			set_bkg_palette( 7, 1, &bar_p[3] );
			set_bkg_palette( 6, 1, &bar_p[3] );
			set_bkg_palette( 5, 1, &bar_p[3] );
			set_bkg_palette( 4, 1, &bar_p[3] );
			set_bkg_palette( 3, 1, &bar_p[3] );
			set_bkg_palette( 2, 1, &bar_p[3] );
			set_bkg_palette( 1, 1, &bar_p[3] );
			set_bkg_palette( 0, 1, &bar_p[3] );
			
			// Select VRAM bank 1 
			VBK_REG = 1;
			// Select VRAM bank 0
			VBK_REG = 0;
			SHOW_BKG;
			DISPLAY_ON;
		}
	}
	else {
		set_bkg_palette( 7, 1, &bar_p[0] );
		set_bkg_palette( 6, 1, &bar_p[0] );
		set_bkg_palette( 5, 1, &bar_p[0] );
		set_bkg_palette( 4, 1, &bar_p[0] );
		set_bkg_palette( 3, 1, &bar_p[0] );
		set_bkg_palette( 2, 1, &bar_p[0] );
		set_bkg_palette( 1, 1, &bar_p[0] );
		set_bkg_palette( 0, 1, &bar_p[0] );
		
		// Select VRAM bank 1 
		VBK_REG = 1;
		// Select VRAM bank 0
		VBK_REG = 0;
		SHOW_BKG;
		
		DISPLAY_ON;
		printf("\n   insideGadgets\nGB Controller v1.2a\n");
		
		// Start button held down at boot, go to options
		if (joypad() & J_START) {
			inOptions = 1;
		}
	}
	
	// B button held down at boot - show key presses
	if (joypad() & J_B) {
		showKeyPresses = 1;
	}
	
	// Normal boot (not going into options), enable the timer
	if (inOptions == 0) { 
		// Setup the timer
		disable_interrupts();
		add_TIM(timer_isr);
		enable_interrupts();
		TMA_REG = 0x60; // Gives about 2ms wake up
		TAC_REG = 0x06;
		set_interrupts(TIM_IFLAG);
	}
	
	while (1) {
		if (inOptions == 1) {
			options();
		}
		else {
			sleep();
			checkInput();
		}
	}
}

void options() {
	if (refreshScreen == 1) {
		gotoxy(0,3);
		printf("  --- Options ---\n");
		printf("  RF Channel: %d  \n", rfChannel);
		printf("  RF Address: %d  \n", rfAddress);
		printf("  Pcks/Refresh: %d  \n", rfPpr);
		printf("\nPress A to transmit\n");
		printf("Power on receiver\n");
		printf("Wait 2-3 seconds\n");
		printf("Power cycle Gameboy\n");
		
		if (transmitOptions == 1) {
			printf("\n\nTransmitting...");
		}
		
		gotoxy(0,arrowLocation+3);
		printf(">");
		
		wait_vbl_done();
		refreshScreen = 0;
	}
	
	// RF channel/address
	if (transmitOptions == 0) {
		if (arrowLocation == 1) {
			if (rfChannel > 1 && (joypad() & J_LEFT)) {
				rfChannel--;
				delay(50);
				refreshScreen = 1;
			}
			if (rfChannel < 125 && (joypad() & J_RIGHT)) {
				rfChannel++;
				delay(50);
				refreshScreen = 1;
			}
		}
		else if (arrowLocation == 2) {
			if (rfAddress > 1 && (joypad() & J_LEFT)) {
				rfAddress--;
				delay(50);
				refreshScreen = 1;
			}
			if (rfAddress < 254 && (joypad() & J_RIGHT)) {
				rfAddress++;
				delay(50);
				refreshScreen = 1;
			}
		}
		else if (arrowLocation == 3) {
			if (rfPpr > 2 && (joypad() & J_LEFT)) {
				rfPpr--;
				delay(50);
				refreshScreen = 1;
			}
			if (rfPpr < 4 && (joypad() & J_RIGHT)) {
				rfPpr++;
				delay(50);
				refreshScreen = 1;
			}
		}
		
		if ((joypad() & J_DOWN) && arrowLocation < 3) {
			arrowLocation++;
			refreshScreen = 1;
			delay(50);
		}
		if ((joypad() & J_UP) && arrowLocation > 1) {
			arrowLocation--;
			refreshScreen = 1;
			delay(50);
		}
	}
	
	// Transmit options
	if (transmitOptions == 0 && (joypad() & J_A)) {
		transmitOptions = 1;
		refreshScreen = 1;
		updatenrf(rfChannel << 8 | rfAddress);
		updateppr(rfPpr);
	}
}

void checkInput() {
	// Blank out first character of last key press
	if (screenOn == 1) {
		gotoxy(0,3);
	}
	
	// Read key pad and set appropriate bits
	kpInput = 0;
	kpRead = joypad();
	if (kpRead & J_A) {
		kpInput |= 0x01;
		if (showKeyPresses == 1 && screenOn == 1) {
			printf("A");
		}
	}
	if (kpRead & J_B) {
		kpInput |= 0x02;
		if (showKeyPresses == 1 && screenOn == 1) {
			printf("B");
		}
	}
	if (kpRead & J_RIGHT) {
		kpInput |= 0x04;
		if (showKeyPresses == 1 && screenOn == 1) {
			printf("R");
		}
	}
	if (kpRead & J_LEFT) {
		kpInput |= 0x08;
		if (showKeyPresses == 1 && screenOn == 1) {
			printf("L");
		}
	}
	if (kpRead & J_UP) {
		kpInput |= 0x10;
		if (showKeyPresses == 1 && screenOn == 1) {
			printf("U");
		}
	}
	if (kpRead & J_DOWN) {
		kpInput |= 0x20;
		if (showKeyPresses == 1 && screenOn == 1) {
			printf("D");
		}
	}
	if (kpRead & J_SELECT) {
		kpInput |= 0x40;
		if (showKeyPresses == 1 && screenOn == 1) {
			printf("T");
		}
	}
	if (kpRead & J_START) {
		kpInput |= 0x80;
		if (showKeyPresses == 1 && screenOn == 1) {
			printf("S");
		}
	}
	
	// Check if the screen is on and then clear out the old key presses / dot shown on the screen 
	if (screenOn == 1) {
		if (showKeyPresses == 1) {
			if (kpInput >= 1) {
				refreshScreen = 1;
				if (kpReadLast != kpInput) {
					printf("      ");
					kpReadLast = kpInput;
				}
			}
			else if (kpInput == 0 && refreshScreen == 1) {
				printf("      ");
				refreshScreen = 0;
			}
		}
		else {
			if (kpInput >= 1) {
				printf(". ");
				refreshScreen = 1;
			}
			else if (kpInput == 0 && refreshScreen == 1) { // Clear characters only once when keys let go
				printf("  ");
				refreshScreen = 0;
			}
		}
	}
	
	// Send the key press data to the ATmega
	senddata(kpInput);
}

void timer_isr(void) { }
